"""
Fundamental constants used throughout Lumos.
"""

#: :math:`3.0 \times 10^{8}`  :math:`\frac{m}{s}`
SPEED_OF_LIGHT = 3.0 * 1e8

#: :math:`532 \times 10^{-9}`  :math:`m`
WAVELENGTH = 532 * 1e-9

#: :math:`6378 \times 10^{3}`  :math:`m`
EARTH_RADIUS = 6378 * 1000

#: :math:`1360`  :math:`\frac{W}{m^2}`
SUN_INTENSITY = 1360